@@ -38,6 +38,7 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 #include <pthread.h>
 #else
 #include <io.h>
+#define unlink _unlink
 #endif
 
 #ifndef OK
@@ -56,7 +57,7 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 # define TRUE 1
 #endif
 
-#define IS_SKYSQL(a) ((a) && strstr((a), "db.skysql.net"))
+#define IS_SKYSQL(a) ((a) && strstr((a), "skysql.mariadb.com"))
 #define SKIP_SKYSQL \
 if (IS_SKYSQL(hostname)) \
 { \
@@ -87,6 +88,7 @@ MYSQL *mysql_default = NULL;  /* default connection */
    ((mysql_default && strstr(mysql_get_server_info(mysql_default), "maxScale")) ||\
     (getenv("srv")!=NULL && (strcmp(getenv("srv"), "maxscale") == 0 ||\
      strcmp(getenv("srv"), "skysql-ha") == 0)))
+
 #define SKIP_MAXSCALE \
 if (IS_MAXSCALE()) \
 { \
@@ -94,9 +96,21 @@ if (IS_MAXSCALE()) \
   return SKIP; \
 }
 
+#define IS_XPAND()\
+   ((mysql_default && strstr(mysql_get_server_info(mysql_default), "Xpand")) ||\
+    (getenv("srv")!=NULL && strcmp(getenv("srv"), "xpand") == 0))
+
+#define SKIP_XPAND \
+if (IS_XPAND()) \
+{ \
+  diag("test disabled with Xpand"); \
+  return SKIP; \
+}
+
 #define SKIP_LOAD_INFILE_DISABLE \
 if (!((mysql->server_capabilities & CLIENT_LOCAL_FILES) &&  \
-         (mysql->options.client_flag & CLIENT_LOCAL_FILES))) { \
+         (mysql->options.client_flag & CLIENT_LOCAL_FILES)) || \
+    IS_XPAND()) { \
   diag("Load local infile not supported"); \
   return SKIP; \
 }
@@ -625,6 +639,11 @@ void get_envvars() {
   }
   if ((envvar= getenv("MYSQL_TEST_PLUGINDIR")))
     plugindir= envvar;
+
+  if (IS_XPAND())
+  {
+
+  }
 }
 
 MYSQL *my_test_connect(MYSQL *mysql,
@@ -663,7 +682,7 @@ void run_tests(struct my_tests_st *test) {
     total++;
   plan(total);
 
-  /* display TLS stats */
+/* display TLS stats */
   mysql= mysql_init(NULL);
   mysql_ssl_set(mysql, NULL, NULL, NULL, NULL, NULL);
 
@@ -749,4 +768,3 @@ void run_tests(struct my_tests_st *test) {
     mysql_close(mysql_default);
   }
 }
-
